import os
import sys

import click
import six
from gevent import pywsgi
from geventwebsocket.handler import WebSocketHandler

from dagster import ExecutionTargetHandle, check
from dagster.cli.load_handle import handle_for_repo_cli_args
from dagster.cli.pipeline import repository_target_argument
from dagster.core.scheduler import SystemCronScheduler
from dagster.core.storage.runs import FilesystemRunStorage, InMemoryRunStorage
from dagster.utils import (
    DEFAULT_REPOSITORY_YAML_FILENAME,
    Features,
    dagster_logs_dir_for_handle,
    dagster_schedule_dir_for_handle,
    is_dagster_home_set,
)

from .app import create_app
from .version import __version__


def create_dagit_cli():
    return ui(auto_envvar_prefix='DAGIT')  # pylint: disable=no-value-for-parameter


REPO_TARGET_WARNING = (
    'Can only use ONE of --repository-yaml/-y, --python-file/-f, --module-name/-m.'
)


@click.command(
    name='ui',
    help=(
        'Run dagit. Loads a repository or pipeline.\n\n{warning}'.format(
            warning=REPO_TARGET_WARNING
        )
        + (
            '\n\n Examples:'
            '\n\n1. dagit'
            '\n\n2. dagit -y path/to/{default_filename}'
            '\n\n3. dagit -f path/to/file.py -n define_repo'
            '\n\n4. dagit -m some_module -n define_repo'
            '\n\n5. dagit -f path/to/file.py -n define_pipeline'
            '\n\n6. dagit -m some_module -n define_pipeline'
            '\n\n7. dagit -p 3333'
            '\n\nOptions Can also provide arguments via environment variables prefixed with DAGIT_'
            '\n\n    DAGIT_PORT=3333 dagit'
        ).format(default_filename=DEFAULT_REPOSITORY_YAML_FILENAME)
    ),
)
@repository_target_argument
@click.option('--host', '-h', type=click.STRING, default='127.0.0.1', help="Host to run server on")
@click.option('--port', '-p', type=click.INT, default=3000, help="Port to run server on")
@click.option('--sync', is_flag=True, help='Use the synchronous execution manager')
@click.option(
    '--log',
    is_flag=True,
    help=(
        'Record logs of pipeline runs. Use --log-dir to specify the directory to record logs to. '
        'By default, logs will be stored under $DAGSTER_HOME.'
    ),
)
@click.option('--log-dir', help="Directory to record logs to", default=None)
@click.option('--schedule-dir', help="Directory to record logs to", default=None)
@click.option(
    '--no-watch',
    is_flag=True,
    help='Disable autoreloading when there are changes to the repo/pipeline being served',
)
@click.version_option(version=__version__, prog_name='dagit')
def ui(host, port, sync, log, log_dir, schedule_dir, no_watch=False, **kwargs):
    handle = handle_for_repo_cli_args(kwargs)

    # add the path for the cwd so imports in dynamically loaded code work correctly
    sys.path.append(os.getcwd())

    if log and not log_dir:
        if not is_dagster_home_set():
            raise click.UsageError(
                '$DAGSTER_HOME is not set and log-dir is not provided. '
                'Set the home directory for dagster by exporting DAGSTER_HOME in your '
                '.bashrc or .bash_profile, or pass in a default directory using the --log-dir flag '
                '\nExamples:'
                '\n  1. export DAGSTER_HOME="~/dagster"'
                '\n  2. --log --log-dir="/dagster_logs"'
            )

        log_dir = dagster_logs_dir_for_handle(handle)

    if Features.SCHEDULER.is_enabled:
        # Don't error if $DAGSTER_HOME is not set
        if not schedule_dir and is_dagster_home_set():
            schedule_dir = dagster_schedule_dir_for_handle(handle)

    check.invariant(
        not no_watch,
        'Do not set no_watch when calling the Dagit Python CLI directly -- this flag is a no-op '
        'at this level and should be set only when invoking dagit/bin/dagit.',
    )
    host_dagit_ui(log, log_dir, schedule_dir, handle, sync, host, port)


def host_dagit_ui(log, log_dir, schedule_dir, handle, use_sync, host, port):
    check.inst_param(handle, 'handle', ExecutionTargetHandle)

    pipeline_run_storage = FilesystemRunStorage(base_dir=log_dir) if log else InMemoryRunStorage()

    if Features.SCHEDULER.is_enabled:
        scheduler = SystemCronScheduler(schedule_dir)
        app = create_app(
            handle, pipeline_run_storage, scheduler, use_synchronous_execution_manager=use_sync
        )
    else:
        app = create_app(handle, pipeline_run_storage, use_synchronous_execution_manager=use_sync)

    server = pywsgi.WSGIServer((host, port), app, handler_class=WebSocketHandler)
    print('Serving on http://{host}:{port}'.format(host=host, port=port))
    try:
        server.serve_forever()
    except OSError as os_error:
        if 'Address already in use' in str(os_error):
            six.raise_from(
                Exception(
                    (
                        'Another process on your machine is already listening on port {port}. '
                        'It is possible that you have another instance of dagit '
                        'running somewhere using the same port. Or it could be another '
                        'random process. Either kill that process or use the -p option to '
                        'select another port.'
                    ).format(port=port)
                ),
                os_error,
            )
        else:
            raise os_error


def main():
    cli = create_dagit_cli()
    # click magic
    cli(obj={})  # pylint:disable=E1120



from __future__ import absolute_import

import os

import nbformat
from dagster_graphql.implementation.context import DagsterGraphQLContext
from dagster_graphql.implementation.pipeline_execution_manager import (
    MultiprocessingExecutionManager,
    SynchronousExecutionManager,
)
from dagster_graphql.schema import create_schema
from flask import Flask, request, send_file, send_from_directory
from flask_cors import CORS
from flask_graphql import GraphQLView
from flask_sockets import Sockets
from graphql.error.base import GraphQLError
from graphql.execution.executors.gevent import GeventExecutor as Executor
from nbconvert import HTMLExporter
from six import text_type

from dagster import ExecutionTargetHandle, check, seven
from dagster.core.scheduler import Scheduler
from dagster.core.storage.runs import RunStorage
from dagster.utils import Features
from dagster.utils.log import get_stack_trace_array

from .subscription_server import DagsterSubscriptionServer
from .templates.playground import TEMPLATE as PLAYGROUND_TEMPLATE
from .version import __version__

try:
    from typing import Any, Dict
except ImportError:
    pass


# based on default_format_error copied and pasted from graphql_server 1.1.1
def format_error_with_stack_trace(error):

    # type: (Exception) -> Dict[str, Any]

    formatted_error = {'message': text_type(error)}  # type: Dict[str, Any]
    if isinstance(error, GraphQLError):
        if error.locations is not None:
            formatted_error['locations'] = [
                {'line': loc.line, 'column': loc.column} for loc in error.locations
            ]
        if error.path is not None:
            formatted_error['path'] = error.path

        # this is what is different about this implementation
        # we print out stack traces to ease debugging
        if hasattr(error, 'original_error') and error.original_error:
            formatted_error['stack_trace'] = get_stack_trace_array(error.original_error)

    return formatted_error


class DagsterGraphQLView(GraphQLView):
    def __init__(self, context, **kwargs):
        super(DagsterGraphQLView, self).__init__(**kwargs)
        self.context = check.inst_param(context, 'context', DagsterGraphQLContext)

    def get_context(self):
        return self.context

    format_error = staticmethod(format_error_with_stack_trace)


def dagster_graphql_subscription_view(subscription_server, context):
    context = check.inst_param(context, 'context', DagsterGraphQLContext)

    def view(ws):
        subscription_server.handle(ws, request_context=context)
        return []

    return view


def static_view(path, file):
    return send_from_directory(
        os.path.join(os.path.dirname(__file__), './webapp/build/static/', path), file
    )


def vendor_view(path, file):
    return send_from_directory(
        os.path.join(os.path.dirname(__file__), './webapp/build/vendor/', path), file
    )


def index_view(_path):
    try:
        return send_file(os.path.join(os.path.dirname(__file__), './webapp/build/index.html'))
    except seven.FileNotFoundError:
        text = '''<p>Can't find webapp files. Probably webapp isn't built. If you are using
        dagit, then probably it's a corrupted installation or a bug. However, if you are
        developing dagit locally, your problem can be fixed as follows:</p>

<pre>cd ./python_modules/
make rebuild_dagit</pre>'''
        return text, 500


def notebook_view(request_args):
    check.dict_param(request_args, 'request_args')

    # This currently provides open access to your file system - the very least we can
    # do is limit it to notebook files until we create a more permanent solution.
    path = request_args['path']
    if not path.endswith('.ipynb'):
        return 'Invalid Path', 400

    with open(os.path.abspath(path)) as f:
        read_data = f.read()
        notebook = nbformat.reads(read_data, as_version=4)
        html_exporter = HTMLExporter()
        html_exporter.template_file = 'basic'
        (body, resources) = html_exporter.from_notebook_node(notebook)
        return '<style>' + resources['inlining']['css'][0] + '</style>' + body, 200


def create_app(
    handle, pipeline_run_storage, scheduler=None, use_synchronous_execution_manager=False
):
    check.inst_param(handle, 'handle', ExecutionTargetHandle)
    check.inst_param(pipeline_run_storage, 'pipeline_run_storage', RunStorage)
    check.opt_inst_param(scheduler, 'scheduler', Scheduler)
    check.bool_param(use_synchronous_execution_manager, 'use_synchronous_execution_manager')

    app = Flask('dagster-ui')
    sockets = Sockets(app)
    app.app_protocol = lambda environ_path_info: 'graphql-ws'

    schema = create_schema()
    subscription_server = DagsterSubscriptionServer(schema=schema)

    if use_synchronous_execution_manager:
        execution_manager = SynchronousExecutionManager()
    else:
        execution_manager = MultiprocessingExecutionManager()

    print('Loading repository...')

    if Features.SCHEDULER.is_enabled:
        context = DagsterGraphQLContext(
            handle=handle,
            execution_manager=execution_manager,
            pipeline_runs=pipeline_run_storage,
            scheduler=scheduler,  # Add schedule argument
            version=__version__,
        )
    else:
        context = DagsterGraphQLContext(
            handle=handle,
            execution_manager=execution_manager,
            pipeline_runs=pipeline_run_storage,
            version=__version__,
        )

    app.add_url_rule(
        '/graphql',
        'graphql',
        DagsterGraphQLView.as_view(
            'graphql',
            schema=schema,
            graphiql=True,
            # XXX(freiksenet): Pass proper ws url
            graphiql_template=PLAYGROUND_TEMPLATE,
            executor=Executor(),
            context=context,
        ),
    )
    sockets.add_url_rule(
        '/graphql', 'graphql', dagster_graphql_subscription_view(subscription_server, context)
    )

    # these routes are specifically for the Dagit UI and are not part of the graphql
    # API that we want other people to consume, so they're separate for now.
    # Also grabbing the magic global request args dict so that notebook_view is testable
    app.add_url_rule('/dagit/notebook', 'notebook', lambda: notebook_view(request.args))

    app.add_url_rule('/static/<path:path>/<string:file>', 'static_view', static_view)
    app.add_url_rule('/vendor/<path:path>/<string:file>', 'vendor_view', vendor_view)
    app.add_url_rule('/<path:_path>', 'index_catchall', index_view)
    app.add_url_rule('/', 'index', index_view, defaults={'_path': ''})

    CORS(app)

    return app



from graphql_ws.constants import GQL_COMPLETE, GQL_DATA
from graphql_ws.gevent import GeventSubscriptionServer, SubscriptionObserver
from rx import Observable


class DagsterSubscriptionServer(GeventSubscriptionServer):
    '''Subscription server that is able to handle non-subscription commands'''

    def __init__(self, middleware=None, **kwargs):
        self.middleware = middleware or []
        super(DagsterSubscriptionServer, self).__init__(**kwargs)

    def execute(self, request_context, params):
        # https://github.com/graphql-python/graphql-ws/issues/7
        params['context_value'] = request_context
        params['middleware'] = self.middleware
        return super(DagsterSubscriptionServer, self).execute(request_context, params)

    def send_execution_result(self, connection_context, op_id, execution_result):
        if execution_result == GQL_COMPLETE:
            return self.send_message(connection_context, op_id, GQL_COMPLETE, {})
        else:
            result = self.execution_result_to_dict(execution_result)
            return self.send_message(connection_context, op_id, GQL_DATA, result)

    def on_start(self, connection_context, op_id, params):
        try:
            execution_result = self.execute(connection_context.request_context, params)
            if not isinstance(execution_result, Observable):
                # pylint cannot find of method
                observable = Observable.of(execution_result, GQL_COMPLETE)  # pylint: disable=E1101
            else:
                observable = execution_result
            observable.subscribe(
                SubscriptionObserver(
                    connection_context,
                    op_id,
                    self.send_execution_result,
                    self.send_error,
                    self.on_close,
                )
            )
        # appropriate to catch all errors here
        except Exception as e:  # pylint: disable=W0703
            self.send_error(connection_context, op_id, str(e))



#!/usr/bin/env python
import os
import signal
import sys
import time

from watchdog.observers import Observer
from watchdog.tricks import AutoRestartTrick


# Use watchdog's python API to auto-restart the dagit-cli process when
# python files in the current directory change. This is a slightly modified
# version of the code in watchdog's `watchmedo` CLI. We've modified the
# KeyboardInterrupt handler below to call handler.stop() before tearing
# down the observer so that repeated Ctrl-C's don't cause the process to
# exit and leave dagit-cli dangling.
#
# Original source:
# https://github.com/gorakhargosh/watchdog/blob/master/src/watchdog/watchmedo.py#L124
#
# Issue:
# https://github.com/gorakhargosh/watchdog/issues/543
class DagsterAutoRestartTrick(AutoRestartTrick):
    def __init__(self, *args, **kwargs):
        super(DagsterAutoRestartTrick, self).__init__(*args, **kwargs)
        self.restarting = False

    def on_any_event(self, event):
        self.restarting = True
        super(DagsterAutoRestartTrick, self).on_any_event(event)
        self.restarting = False


def handle_sigterm(_signum, _frame):
    raise KeyboardInterrupt()


def main():
    # Build the dagit-cli command, omitting the --no-watch arg if present
    watch = True
    command = ['dagit-cli']
    for arg in sys.argv[1:]:
        if arg == '--no-watch':
            watch = False
        elif arg == '--help':
            watch = False
            command.append(arg)
        elif arg == '--version':
            watch = False
            command.append(arg)
        else:
            command.append(arg)

    # If not using watch mode, just call the command
    if not watch:
        os.execvp(command[0], command)

    signal.signal(signal.SIGTERM, handle_sigterm)

    handler = DagsterAutoRestartTrick(
        command=command,
        patterns=['*.py'],
        ignore_patterns=[],
        ignore_directories=[],
        stop_signal=signal.SIGINT,
        kill_after=0.5,
    )
    handler.start()

    print('Will be watching for file changes...')
    observer = Observer(timeout=1)
    observer.schedule(handler, '.', True)
    observer.start()

    try:
        while not handler.process or handler.restarting or handler.process.poll() is None:
            time.sleep(1)
    except KeyboardInterrupt:
        pass

    handler.stop()
    observer.stop()
    observer.join()
